@@ -1,8 +1,11 @@
 import os
+import pathlib
 from copy import deepcopy
+from decimal import Decimal
 
 import numpy as np
 from scipy.ndimage.interpolation import map_coordinates, zoom
+from skimage.transform import resize_local_mean, downscale_local_mean
 
 from hdrio import imread
 
@@ -21,14 +24,6 @@ SUPPORTED_FORMATS = [
     'cube',
 ]
 
-ROTATION_FORMATS = [
-    'DCM',
-    'EA###',    # TODO
-    'EV',       # TODO
-    'Q',        # TODO
-]
-
-
 #From Dan:
 #  I've generated these using the monochromatic albedo values from here:
 #  http://agsys.cra-cin.it/tools/solarradiation/help/Albedo.html
@@ -43,23 +38,26 @@ GROUND_ALBEDOS = {
 }
 
 
+def isPath(var):
+    return isinstance(var, str) or isinstance(var, pathlib.PurePath)
+
+
 class EnvironmentMap:
-    def __init__(self, im, format_=None, copy=True, color=True):
+    def __init__(self, im, format_=None, copy=True, channels=3):
         """
         Creates an EnvironmentMap.
 
-        :param im: Image path or data to be converted to an EnvironmentMap, or 
-                   the height of an empty EnvironmentMap.
-        :param format_: EnvironmentMap format. Can be `Angular`, ...
+        :param im: Image path (str, pathlib.Path) or data (np.ndarray) representing 
+                   an EnvironmentMap, or the height (int) of an empty EnvironmentMap.
+        :param format_: EnvironmentMap format. Can be any from SUPPORTED_FORMATS.
         :param copy: When a numpy array is given, should it be copied.
-        :param color: When providing an integer, create an empty color or
-                      grayscale EnvironmentMap.
-        :type im: float, numpy array
-        :type format_: string
+        :param channels: Number of channels (e.g., 1=grayscale, 3=color).
+        :type im: str, Path, int, np.ndarray
+        :type format_: str
         :type copy: bool
         """
-        if not format_ and isinstance(im, str):
-            filename = os.path.splitext(im)[0]
+        if not format_ and isPath(im):
+            filename = os.path.splitext(str(im))[0]
             metadata = EnvmapXMLParser("{}.meta.xml".format(filename))
             format_ = metadata.getFormat()
 
@@ -70,31 +68,19 @@ class EnvironmentMap:
 
         self.format_ = format_.lower()
 
-        if isinstance(im, str):
+        if isPath(im):
             # We received the filename
-            self.data = imread(im)
+            self.data = imread(str(im))
         elif isinstance(im, int):
             # We received a single scalar
             if self.format_ == 'latlong':
-                if color:
-                    self.data = np.zeros((im, im*2, 3))
-                else:
-                    self.data = np.zeros((im, im*2))
+                self.data = np.zeros((im, im*2, channels))
             elif self.format_ == 'skylatlong':
-                if color:
-                    self.data = np.zeros((im, im*4, 3))
-                else:
-                    self.data = np.zeros((im, im*4))
+                    self.data = np.zeros((im, im*4, channels))
             elif self.format_ == 'cube':
-                if color:
-                    self.data = np.zeros((im, round(3/4*im), 3))
-                else:
-                    self.data = np.zeros((im, round(3/4*im)))
+                self.data = np.zeros((im, round(3/4*im), channels))
             else:
-                if color:
-                    self.data = np.zeros((im, im, 3))
-                else:
-                    self.data = np.zeros((im, im))
+                    self.data = np.zeros((im, im, channels))
         elif type(im).__module__ == np.__name__:
             # We received a numpy array
             self.data = np.asarray(im, dtype='double')
@@ -103,11 +89,14 @@ class EnvironmentMap:
                 self.data = self.data.copy()
         else:
             raise Exception('Could not understand input. Please provide a '
-                            'filename, a single size (height) or an image.')
+                            'filename (str), an height (integer) or an image '
+                            '(np.ndarray).')
 
-        self.backgroundColor = np.array([0, 0, 0]) if self.data.ndim == 2 else np.zeros(self.data.shape[-1])
+        self.backgroundColor = np.zeros(self.data.shape[-1])
+        self.validate()
 
-        # Ensure size is valid
+    def validate(self):
+        # Ensure the envmap is valid
         if self.format_ in ['sphere', 'angular', 'skysphere', 'skyangular']:
             assert self.data.shape[0] == self.data.shape[1], (
                 "Sphere/Angular formats must have the same width/height")
@@ -117,6 +106,8 @@ class EnvironmentMap:
         elif self.format_ == 'skylatlong':
             assert 4*self.data.shape[0] == self.data.shape[1], (
                 "SkyLatLong format width should be four times the height")
+        
+        assert self.data.ndim == 3, "Expected 3-dim array. For grayscale, use [h,w,1]."
 
     @classmethod
     def fromSkybox(cls, top, bottom, left, right, front, back):
@@ -235,14 +226,12 @@ class EnvironmentMap:
 
     def interpolate(self, u, v, valid=None, order=1, filter=True):
         """"Interpolate to get the desired pixel values."""
+        u = u.copy()
+        v = v.copy()
+        
         # Repeat the first and last rows/columns for interpolation purposes
-        if len(self.data.shape) == 2:
-            h, w = self.data.shape
-            d = 0
-            source = np.empty((h + 2, w + 2))
-        else:
-            h, w, d = self.data.shape
-            source = np.empty((h + 2, w + 2, d))
+        h, w, d = self.data.shape
+        source = np.empty((h + 2, w + 2, d))
 
         source[1:-1, 1:-1] = self.data
         source[0,1:-1] = self.data[0,:]; source[0,0] = self.data[0,0]; source[0,-1] = self.data[0,-1]
@@ -251,41 +240,37 @@ class EnvironmentMap:
         source[1:-1,-1] = self.data[:,-1]
 
         # To avoid displacement due to the padding
-        u += 1./self.data.shape[1]
-        v += 1./self.data.shape[0]
+        u += 0.5/self.data.shape[1]
+        v += 0.5/self.data.shape[0]
         target = np.vstack((v.flatten()*self.data.shape[0], u.flatten()*self.data.shape[1]))
 
-        if d == 0:
-            data = np.zeros((u.shape[0], u.shape[1]))
-            map_coordinates(source, target, output=data.ravel(), cval=np.nan, order=order, prefilter=filter)
-        else:
-            data = np.zeros((u.shape[0], u.shape[1], d))
-            for c in range(d):
-                map_coordinates(source[:,:,c], target, output=data[:,:,c].reshape(-1), cval=np.nan, order=order, prefilter=filter)
+        data = np.zeros((u.shape[0], u.shape[1], d))
+        for c in range(d):
+            map_coordinates(source[:,:,c], target, output=data[:,:,c].reshape(-1), cval=np.nan, order=order, prefilter=filter)
         self.data = data
 
-        # In original: valid &= ~isnan(data)...
-        # I haven't included it here because it may mask potential problems...
         if valid is not None:
             self.setBackgroundColor(self.backgroundColor, valid)
 
         return self
 
-    def setBackgroundColor(self, color, valid):
+    def setBackgroundColor(self, color, valid=None):
         """Sets the area defined by ~valid to color."""
+        if valid is None:
+            _, _, _, valid = self.worldCoordinates()
+
         assert valid.dtype == 'bool', "`valid` must be a boolean array."
         assert valid.shape[:2] == self.data.shape[:2], "`valid` must be the same size as the EnvironmentMap."
 
         self.backgroundColor = np.asarray(color)
+        if self.backgroundColor.size == 1 and self.data.shape[2] != self.backgroundColor.size:
+            self.backgroundColor = np.tile(self.backgroundColor, (self.data.shape[2],))
 
-        grayscale = len(self.data.shape) == 2
+        assert self.data.shape[2] == self.backgroundColor.size, "Channel number mismatch when setting background color"
 
-        if grayscale:
-            self.data[:,:][np.invert(valid)] = self.backgroundColor.dot(np.array([0.299, 0.587, 0.114]).T)
-        else:
-            nb_channels = self.data.shape[2]
-            for c in range(nb_channels):
-                self.data[:,:,c][np.invert(valid)] = self.backgroundColor[c]
+        mask = np.invert(valid)
+        if mask.sum() > 0:
+            self.data[np.tile(mask[:,:,None], (1, 1, self.data.shape[2]))] = np.tile(self.backgroundColor, (mask.sum(),))
 
         return self
 
@@ -299,6 +284,8 @@ class EnvironmentMap:
         :type targetFormat: integer
 
         """
+        self.validate()
+
         assert targetFormat.lower() in SUPPORTED_FORMATS, (
             "Unknown format: {}".format(targetFormat))
 
@@ -314,17 +301,18 @@ class EnvironmentMap:
 
         return self
 
-    def rotate(self, format, input_):
+    def rotate(self, dcm):
         """
         Rotate the environment map.
 
-        :param format: Rotation type
         :param input: Rotation information (currently only 3x3 numpy matrix)
         """
-        assert format.upper() in ROTATION_FORMATS, "Unknown rotation type '{}'".format(format)
+        self.validate()
+
+        assert type(dcm).__module__ == np.__name__ and dcm.ndim == 2 and dcm.shape == (3, 3)
         dx, dy, dz, valid = self.worldCoordinates()
 
-        ptR = np.dot(input_, np.vstack((dx.flatten(), dy.flatten(), dz.flatten())))
+        ptR = np.dot(dcm, np.vstack((dx.flatten(), dy.flatten(), dz.flatten())))
         dx, dy, dz = ptR[0].reshape(dx.shape), ptR[1].reshape(dy.shape), ptR[2].reshape(dz.shape)
 
         dx = np.clip(dx, -1, 1)
@@ -336,14 +324,16 @@ class EnvironmentMap:
 
         return self
 
-    def resize(self, targetSize, order=1):
+    def resize(self, targetSize, order=1, debug=False):
         """
         Resize the current environnement map to targetSize.
-        targetSize can be a tuple or a single number, in which case the same factor is assumed
-        for both u and v.
-        If 0 < targetSize < 1, treat it as a ratio
-        If targetSize > 1, treat it as new dimensions to use
+        The energy-preserving "pixel mixing" alrogithm is used when downscaling.
+
+        `targetSize` is either the desired `height` or `(height, width)`.
+        `order` is only used when upsampling.
         """
+        self.validate()
+
         if not isinstance(targetSize, tuple):
             if self.format_ == 'latlong':
                 targetSize = (targetSize, 2*targetSize)
@@ -353,48 +343,120 @@ class EnvironmentMap:
                 targetSize = (targetSize, round(3/4*targetSize))
             else:
                 targetSize = (targetSize, targetSize)
+        
+        # downsampling
+        if targetSize[0] < self.data.shape[0]:
 
+            if debug == True:
+                old_mean = self.data.mean()
+
+            # check if integer
+            if (Decimal(self.data.shape[0]) / Decimal(targetSize[0])) % 1 == 0:
+                if debug is True:
+                    print("integer resize")
+                fac = self.data.shape[0] // targetSize[0]
+                self.data = downscale_local_mean(self.data, (fac, fac, 1))
+            else:    
+                self.data = resize_local_mean(self.data, targetSize, grid_mode=True, preserve_range=True)
+
+            if debug == True:
+                print("Energy difference in resize: {:.04f}".format(self.data.mean()/old_mean - 1.))
+
+            return self
+        
+        # upsampling
         _size = []
         for i in range(2):
             _size.append(targetSize[i] / self.data.shape[i] if targetSize[i] > 1. else targetSize[i])
 
-        if len(self.data.shape) > 2:
-            _size.append(1.0)   # To ensure we do not "scale" the color axis...
+        _size.append(1.0)
 
         self.data = zoom(self.data, _size, order=order)
         return self
 
-    def toIntensity(self):
+    def toIntensity(self, mode="ITU BT.709", colorspace="linear"):
         """
-        Returns intensity-version of the environment map.
-        This function assumes the CCIR 601 standard to perform internsity conversion.
+        Converts the environment map to grayscale.
+
+        `mode` can be either :
+               "ITU BT.601" (luma in sRGB),
+               "ITU BT.709" (assumes the envmap is linear),
+               mean (mean of the channels)
+        `colorspace`: either "sRGB" or "linear"
         """
-        assert len(self.data.shape) == 3, "Data should have exactly 3 dimensions"
+        self.validate()
 
         if self.data.shape[2] != 3:
-            print("Envmap doesn't have 3 channels. This function won't do anything.")
-        else:
-            self.data = 0.299 * self.data[...,0] + 0.587 * self.data[...,1] + 0.114 * self.data[...,2]
+            print("Envmap does not have 3 channels. This function won't do anything.")
+            return self
+        
+        if colorspace.lower() == "sRGB":
+            if self.data.max() > 1.:
+                raise Exception("Error during sRGB to linear conversion: data is > 1. Please linearize "
+                                "the data beforehand or normalize it [0, 1].")
+            self.data = self.data**2.2
+
+        if mode == "ITU BT.601":
+            self.data = 0.299*self.data[...,0] + 0.587*self.data[...,1] + 0.114*self.data[...,2]
             self.data = self.data[:,:,np.newaxis]
+        elif mode == "ITU BT.709":
+            self.data = 0.2126*self.data[...,0] + 0.7152*self.data[...,1] + 0.0722*self.data[...,2]
+            self.data = self.data[:,:,np.newaxis]
+        elif mode == "mean":
+            self.data = np.mean(self.data, axis=2, keepdims=True)
+
+        if colorspace.lower() == "sRGB":
+            self.data = self.data**(1./2.2)
+
         return self
 
-    def setHemisphereAlbedo(self, normal, value):
-        """Sets an whole hemisphere defined by `normal` to a given `value`.
-        Useful to set the ground albedo."""
-        raise NotImplementedError()
+    def getHemisphere(self, normal, channels=True):
+        """
+        
+        normal: 
+        """
+        normal = np.asarray(normal, dtype=np.float32).reshape((-1))
+        assert normal.size == 3, "unknown normal shape, should have 3 elements"
+        normal /= np.linalg.norm(normal)
+
+        x, y, z, _ = self.worldCoordinates()
+        xyz = np.dstack((x, y, z))
+
+        mask = xyz.dot(normal) > 0
+        if channels == False:
+            return mask
+        
+        return np.tile(mask[:,:,None], (1, 1, self.data.shape[2]))
+
+    def setHemisphereValue(self, normal, value):
+        """Sets an whole hemisphere defined by `normal` to a given `value`."""
+
+        mask = self.getHemisphere(normal, channels=True)
+
+        value = np.asarray(value)
+        assert value.size in (self.data.shape[2], 1), ("Cannot understand value: should "
+            "be size 1 or channels ({})".format(self.data.shape[2]))
+
+        if value.size == 1 and self.data.shape[2] != value.size:
+            value = np.tile(value, (self.data.shape[2],))
+
+        self.data[mask] = np.tile(value, (mask.sum()//self.data.shape[2],))
+
+        return self
 
     def getMeanLightVectors(self, normals):
         """Compute the mean light vector of the environment map for the normals given.
         Normals should be 3xN.
         Output is 3xN.
         """
+        self.validate()
+
         normals = np.asarray(normals)
+        normals /= np.linalg.norm(normals, 1)
         solidAngles = self.solidAngles()
         solidAngles /= np.nansum(solidAngles) # Normalize to 1
-        normals /= np.linalg.norm(normals, 1)
 
         x, y, z, _ = self.worldCoordinates()
-
         xyz = np.dstack((x, y, z))
 
         visibility = xyz.dot(normals) > 0
@@ -412,6 +474,8 @@ class EnvironmentMap:
         :rotation_matrix: Camera rotation matrix.
         :image: The image to project."""
 
+        self.validate()
+
         targetDim = self.data.shape[0]
         targetFormat = self.format_
 
@@ -466,12 +530,14 @@ class EnvironmentMap:
                "normal+uv": returns (crop, u, v), where (u,v) are the coordinates
                             of the crop."""
 
+        self.validate()
+
         coords = self._cameraCoordinates(vfov, rotation_matrix, ar,
                                          resolution, projection, mode)
 
         if mode == "mask":
             return coords
-        
+
         target = self.copy()
         if target.format_ != "latlong":
             target = target.convertTo("LatLong")
@@ -515,8 +581,8 @@ class EnvironmentMap:
             dmask = z < 0
             mask = hmask & vmask & dmask
 
-            e = EnvironmentMap(mask[:,:,np.newaxis], 'LatLong')
-            e.rotate("DCM", rotation_matrix)
+            e = EnvironmentMap(mask[:,:,np.newaxis], self.format_)
+            e.rotate(rotation_matrix)
 
             mask = e.data[:,:,0]
             return mask
@@ -553,22 +619,9 @@ def rotation_matrix(azimuth, elevation, roll=0):
     :roll: counter-clockwise (negative) to clockwise (positive) [rad]"""
     return rotz(roll).dot(rotx(elevation)).dot(roty(-azimuth))
 
+
 def downscaleEnvmap(nenvmap, sao, sat, times):
-    """Energy-preserving environment map downscaling by factors of 2.
-    Usage:
-    sao = EnvironmentMap(512, 'LatLong').solidAngles() # Source envmap solid angles, could be replaced by `sao = envmap.soldAngles()`
-    sat = EnvironmentMap(128, 'LatLong').solidAngles() # Target envmap solid angles
-    downscaleEnvmap(envmap, sao, sat, 3)
-    Note : `times` is the number of downscales, so the total downscaling factor is 2**times"""
-    nenvmap.data *= sao[:, :, np.newaxis]
-    nenvmap.data = np.pad(nenvmap.data, [(0, 1), (0, 1), (0, 0)], 'constant')
-    sx = np.cumsum(nenvmap.data, axis=1)
-    tmp = sx[:, 2**times::2**times, ...] - sx[:, :-2**times:2**times, ...]
-    sy = np.cumsum(tmp, axis=0)
-    nenvmap.data = sy[2**times::2**times, :, ...] - sy[:-2**times:2**times, :, ...]
-    if nenvmap.data.shape[1] > 2*nenvmap.data.shape[0]:
-        nenvmap.data[:, -2, :] += nenvmap[:, -1, :]
-        nenvmap.data = nenvmap.data[:, :-1, :]
-    nenvmap.data /= sat[:, :, np.newaxis]
-    nenvmap.data = np.ascontiguousarray(nenvmap.data)
-    return nenvmap
+    """Deprecated"""
+    print("This function is deprecated and will be removed in an ulterior version of skylibs.")
+    sz = sat.shape[0]
+    return nenvmap.resize(sz)
